-- Filename:     microprocessor.vhd
-- Filetype:     VHDL code
-- Date:         23 may 2012
-- Update:       26 may 2015
-- Description:  Description of a simle microprocessor
-- Author:       J. op den Brouw
-- State:        demo
-- Error:        -
-- Version:      1.1.2
-- Copyright:    (c)2015, De Haagse Hogeschool
--
-- This is VHDL description of a simple microprocessor.
-- It's a demonstration to show that it is not hard to
-- describe such a system.

-- The package with common descriptions for the processor system.
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

package microprocessor_common is
    -- The data size and types
    constant data_size : positive := 8;
    subtype data_type is unsigned (7 downto 0);
    subtype data_type1 is unsigned (8 downto 0);

    constant data_zeros : data_type := (others => '0');
	 
    -- ALU operations (ops) is a 4-bit vector. It instructs
	 -- the ALU to perform the desired operation.
    subtype alu_ops_type is std_logic_vector (3 downto 0);
    constant alu_nop   : alu_ops_type := "0000";	 
    constant alu_passb : alu_ops_type := "0001";
    constant alu_add   : alu_ops_type := "0010";
    constant alu_sub   : alu_ops_type := "0011";
    constant alu_and   : alu_ops_type := "0100";
    constant alu_or    : alu_ops_type := "0101";
    constant alu_exor  : alu_ops_type := "0110";
    constant alu_notb  : alu_ops_type := "0111";
    constant alu_shlb  : alu_ops_type := "1000";	 
    constant alu_shrb  : alu_ops_type := "1001";	 
    constant alu_rolb  : alu_ops_type := "1010";	 
    constant alu_rorb  : alu_ops_type := "1011";
    constant alu_wrfb  : alu_ops_type := "1100";	 
    constant alu_rdf   : alu_ops_type := "1101";	 
	 -- 1110 not used
    -- 1111 not used

    -- JUMP operations (ops) is a 3-bit vector. It instructs
	 -- the jump checker to perform the desired operation.
    subtype jump_ops_type is std_logic_vector (2 downto 0);
    constant jump_jn   : jump_ops_type := "000";
    constant jump_jnz  : jump_ops_type := "001";
    constant jump_jz   : jump_ops_type := "010";
    constant jump_jnc  : jump_ops_type := "011";
    constant jump_jc   : jump_ops_type := "100";
    constant jump_jncz : jump_ops_type := "101";
    constant jump_jcz  : jump_ops_type := "110";
    constant jump_j    : jump_ops_type := "111";
    -- alternative set
    -- jn is same
    constant jump_jne  : jump_ops_type := "001";
    constant jump_je   : jump_ops_type := "010";
    constant jump_jge  : jump_ops_type := "011";
    constant jump_jl   : jump_ops_type := "100";
    constant jump_jg   : jump_ops_type := "101";
    constant jump_jle  : jump_ops_type := "110";
    constant jump_jmp  : jump_ops_type := "111";

    -- Address to jumo to.
    constant address_size : positive := 128;
    subtype  address_type is unsigned (6 downto 0);
    constant address_zeros : address_type := (others => '0');

    -- Full format of an instruction. An instruction is 30 bits wide.
    -- A number of microprocessor-common instructions are defined here.
    constant instr_size : positive := 30;	 
    subtype  instr_type is std_logic_vector (29 downto 0);
    -- Full no-operations
    constant instr_nop           : instr_type := (others => '0');
    constant instr_all           : instr_type := (others => '1');
	 constant instr_undef         : instr_type := (others => '-');
    constant instr_test          : instr_type := ("111000100011111111110000000000");
    constant instr_jmp0          : instr_type := ("000000000000000000001110000000");
    -- Instruction definitions
    constant instr_ld_r0_c0a     : instr_type := ("--101000001000010101000-------");
    constant instr_ld_r0_c01     : instr_type := ("--101000001000000011000-------");
    constant instr_ld_r0_c08     : instr_type := ("--101000001000010001000-------");
    constant instr_ld_r0_c0d     : instr_type := ("--101000001000011011000-------");
	 constant instr_ld_r0_r2      : instr_type := ("--010000001--------1000-------");
	 constant instr_ld_r0_r3      : instr_type := ("--011000001--------1000-------");
    constant instr_ld_r1_c07     : instr_type := ("--101010001000001111000-------");
    constant instr_ld_r1_c08     : instr_type := ("--101010001000010001000-------");
    constant instr_ld_r1_caa     : instr_type := ("--101010001101010101000-------");
    constant instr_ld_r1_c0b     : instr_type := ("--101010001000010111000-------");
    constant instr_ld_r1_cff     : instr_type := ("--101010001111111111000-------");
    constant instr_ld_r2_cff     : instr_type := ("--101100001111111111000-------");
    constant instr_ld_r2_caf     : instr_type := ("--101100001101011111000-------");
    constant instr_ld_r2_c00     : instr_type := ("--101100001000000001000-------");
	 constant instr_ld_r2_in      : instr_type := ("--100100001--------1000-------");
    constant instr_ld_r3_c00     : instr_type := ("--101110001000000001000-------");
    constant instr_ld_r3_c2f     : instr_type := ("--101110001001011111000-------");
    constant instr_ld_r3_cff     : instr_type := ("--101110001111111111000-------");
    constant instr_ld_r3_c7f     : instr_type := ("--101110001011111111000-------");
    constant instr_ld_r3_c3f     : instr_type := ("--101110001001111111000-------");
    constant instr_ld_r3_r2      : instr_type := ("--011100001--------1000-------");
    constant instr_cmp_r0_c00    : instr_type := ("00101--0011000000000000-------");
    constant instr_cmp_r1_c00    : instr_type := ("01101--0011000000000000-------");
    constant instr_je_a0c        : instr_type := ("-------00000000000000100001100");
    constant instr_je_a0b        : instr_type := ("-------00000000000000100001011");
    constant instr_jge_a03       : instr_type := ("-------00000000000000110000011");
    constant instr_jne_a03       : instr_type := ("-------00000000000000010000011");
    constant instr_ld_f_c00      : instr_type := ("00101001100------000000-------");
    constant instr_sub_r0_r0_c01 : instr_type := ("00101000011000000011000-------");
    constant instr_sub_r1_r1_c01 : instr_type := ("01101010011000000011000-------");
    constant instr_sub_r2_r2_c00 : instr_type := ("10101100011000000001000-------");
    constant instr_sub_r3_r3_c00 : instr_type := ("11101110011000000001000-------");
    constant instr_add_r2_r2_r0  : instr_type := ("00010100010--------1000-------");
    constant instr_add_r3_r3_c00 : instr_type := ("11101110010000000001000-------");
	 constant instr_shr_r2_r2     : instr_type := ("--010101001--------1000-------");
	 constant instr_ror_r0_r0     : instr_type := ("--000001011--------1000-------");
    constant instr_or_r0_r0_c80  : instr_type := ("00101000101100000001000-------");
    constant instr_jmp_a04       : instr_type := ("-------0000--------01110000100");
    constant instr_jmp_a01       : instr_type := ("-------0000--------01110000001");
    constant instr_jmp_a0c       : instr_type := ("-------00000000000001110001100");
    constant instr_jmp_a09       : instr_type := ("-------00000000000001110001001");
    constant instr_jmp_a0b       : instr_type := ("-------00000000000001110001011");
    constant instr_jnc_a05       : instr_type := ("-------00000000000000110000101");
    constant instr_jnc_a09       : instr_type := ("-------00000000000000110001001");

    -- The program is held in an 128-instruction word, 30 bit wide ROM.	 
    type  rom_table is array (0 to address_size-1) of instr_type;
	 
	 -- Program starts here
    constant rom : rom_table := (
        -- Just some instructions to test...
--        instr_ld_r0_c0a,
--        instr_ld_r1_caa,
--        instr_ld_r2_cff,
--        instr_ld_r3_c00,
--        instr_cmp_r0_c00,
--        instr_je_a0c,
--        instr_ld_f_c00,
--        instr_sub_r1_r1_c01,
--        instr_sub_r0_r0_c01,
--        instr_jmp_a04,
--        instr_nop,
--        instr_nop,
--        instr_jmp_a0c,
		  -- The next program performs a multiplication of the contents of R0
		  -- with R1 and the 16-bit result is placed in R3 (high) and R2 (low).
		  -- The lower result byte is copied to R0.
		  -- It uses the simple successive add algorithm.
        instr_ld_r0_c0d,     -- Load R0 with constant 0D hex
        instr_ld_r1_c0b,     -- Load R1 with constant 0B hex
        instr_ld_r2_c00,     -- Load R2 with constant 00 hex
        instr_ld_r3_r2,      -- Load R2 with contents of R3 (move)
        instr_cmp_r1_c00,    -- Compare R1 with constant 00 hex
        instr_je_a0b,        -- Jump if equal (R1 == 0) to address 0B hex
        instr_ld_f_c00,      -- Load flags with zeros
        instr_add_r2_r2_r0,  -- Add:  R2 := R2 + R0 (+carry)
        instr_add_r3_r3_c00, -- Add:  R3 := R3 + 0 + carry
        instr_sub_r1_r1_c01, -- Sub:  R1 := R1 - 1 (carry is 0)
        instr_jmp_a04,       -- Jump to address 04 hex
        instr_ld_r0_r2,      -- Load R0 with contents of R2
        instr_jmp_a0c,       -- Jump to address 0C hex (itself)
		  -- The next program determines the number of 1's in one sample
		  -- of the input
--		  instr_ld_r3_c00,     -- Clear result
--		  instr_ld_r1_c08,     -- Counter to 8
--		  instr_ld_r2_in,      -- Input data to R2
--		  instr_shr_r2_r2,     -- Shift right R2 
--		  instr_add_r3_r3_c00, -- Add carry if any
--		  instr_sub_r1_r1_c01, -- Subtract 1 of counter
--		  instr_cmp_r1_c00,    -- Counter reached 0?
--		  instr_jne_a03,       -- Loop back
--		  instr_ld_r0_r3,      -- R3 to R0 for output
--		  instr_jmp_a09,       -- Hold
        -- new program
--		  instr_ld_r0_c0a,     -- load
--		  instr_sub_r0_r0_c01,
--		  instr_sub_r0_r0_c01,
--		  instr_sub_r0_r0_c01,
--		  instr_sub_r0_r0_c01,
--		  instr_sub_r0_r0_c01,
--		  instr_sub_r0_r0_c01,
--		  instr_sub_r0_r0_c01,
--		  instr_sub_r0_r0_c01,
--		  instr_sub_r0_r0_c01,
--		  instr_sub_r0_r0_c01,
--		  instr_sub_r0_r0_c01,
--		  instr_sub_r0_r0_c01,
--		  instr_sub_r0_r0_c01,
--		  instr_sub_r0_r0_c01,
--		  instr_sub_r0_r0_c01,
        -- This program rotates a single bit right endlessly.
		  -- It uses a delay routine to tick away cpu time.
		  -- See http://dotasm.com/sample-code/103-delays/ for info
		  -- on the delay loop
		  -- Start program..
--		  instr_ld_r0_c08,     -- Load R0 with constant 1
--		  instr_ld_f_c00,      -- Clear C and Z
--		  instr_ror_r0_r0,     -- Rotate R0 once
--		  instr_jnc_a05,       -- There's no carry so jump
--		  instr_or_r0_r0_c80,  -- Set R0 to 80h (carry in high bit)
--		  -- Init the delay loop counters...
--		  instr_ld_r1_c07,     -- Load the delay routine with 2faf07h
--		  instr_ld_r2_caf,     -- Which results in a 0.25 sec delay.
--		  instr_ld_r3_c2f,     --
--		  instr_ld_f_c00,      -- Clear C and Z
--		  -- Do the delay... 4 clock ticks per loop...
--		  instr_sub_r1_r1_c01, -- subtract 1 from R1 (carry = 0)
--		  instr_sub_r2_r2_c00, -- sub carry  from R2
--		  instr_sub_r3_r3_c00, -- sub carry  from R3
--		  instr_jnc_a09,       -- jump if no carry to address 7 (sub)
--		  -- Delay done, so repeat the rotate process...
--		  instr_jmp_a01,       -- Jump to address 2 (clear flags)
--		  -- End of program, fill with nop's or whatever...
--        instr_jmp0,
--		  instr_test,
		  others => instr_undef);
end package microprocessor_common;


-- The microprocessor has a clock, a asynchronous reset,
-- an 8-bit input and an 8-bit output (otherwise the synthesizer has
-- nothing to do...).
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.microprocessor_common.all;

entity microprocessor is
    port (clk   : in std_logic;
          areset : in std_logic;
          input : in data_type;
          output: out data_type
         );
end entity microprocessor;

architecture rtl of microprocessor is
-- All internal signals
signal zupdate, cupdate, zflag, cflag : std_logic;
signal muxaout, muxbout, alu_result : data_type;
signal acode : alu_ops_type;
signal jcode : jump_ops_type;
signal ldpc : std_logic;
signal reg0, reg1, reg2, reg3 : data_type;
signal enr0, enr1, enr2, enr3 : std_logic;
signal sela, seld : std_logic_vector (1 downto 0);
signal selb : std_logic_vector (2 downto 0);
signal const : data_type;
signal en_d : std_logic;
signal address, pc : address_type;
begin

    -- The Zero and Carry flags. They are really flip-flops and
    -- depend on the outcome of the ALU result.
    flags : process (clk, areset) is
    begin
        if areset = '1' then
            cflag <= '0';
            zflag <= '0';
        elsif rising_edge(clk) then
            cflag <= cupdate;
            zflag <= zupdate;
        end if;				
    end process flags;

    -- The ALU. This is a pure combinational circuit that can
	 -- add, subtract, do logic, shifts and rotates and more.
    -- Some ALU ops use the flags, some don't. Note that the
    -- behavior of the flags depends on the ALU op.
    alu : process (acode, muxaout, muxbout, zflag, cflag) is
    variable alu_result_int : data_type1;
    variable z_int, c_int : std_logic;
    begin
        -- Set result to 0 and get the current flag data
        alu_result_int := (others => '0');
        z_int := zflag;
        c_int := cflag;
        -- Do the operation
        case acode is
            when alu_nop =>
                alu_result_int := '0' & muxbout;
            when alu_passb =>
                alu_result_int := '0' & muxbout;
            when alu_add =>
                alu_result_int := ('0' & muxaout) + ('0' & muxbout) + ("" & cflag);
                c_int := alu_result_int(8);
            when alu_sub =>
                --alu_result_int := ('0' & muxaout) - ('0' & muxbout) - ("" & cflag);
                --c_int := alu_result_int(8);
					 -- Subtraction can be performed using two's complement even if the numbers
					 -- are all unsigned. 
					 -- Two's complement subtraction: A, B: 8 bit un/signed, Cin : 1 bit Carry In
					 --     A - B - Cin  ==  A + (not(B) + 1) - Cin  ==  A + not(B) + (1-Cin)
					 --                  ==  A + not(B) + not(Cin)
					 --     Cout = not(Carry out of the adder hardware)
                alu_result_int := ('0' & muxaout) + ('0' & not muxbout) + ("" & not cflag);
                c_int := not alu_result_int(8);
            when alu_and =>
                alu_result_int := '0' & (muxaout and muxbout);
            when alu_or =>
                alu_result_int := '0' & (muxaout or muxbout);
            when alu_exor =>
                alu_result_int := '0' & (muxaout xor muxbout);
            when alu_notb =>
                alu_result_int := '0' & (not muxbout);
            when alu_shlb =>
                alu_result_int := muxbout & '0';
                c_int := alu_result_int(8);
            when alu_shrb =>
					 -- we need a little trick here...
					 alu_result_int := "00" & muxbout(7 downto 1);
                c_int := muxbout(0);
            when alu_rolb =>
                alu_result_int := muxbout & cflag;
                c_int := alu_result_int(8);
            when alu_rorb =>
                -- we need a little trick here...
                alu_result_int := '0' & cflag & muxbout(7 downto 1);
                c_int := muxbout(0);
            when alu_wrfb =>
                alu_result_int := ('0' & muxbout);
                c_int := alu_result_int(0);
                z_int := alu_result_int(1);
            when alu_rdf =>
                alu_result_int := (0 => cflag, 1 => zflag, others => '0');
            when others =>
                alu_result_int := ('0' & muxbout);
       end case;
        -- Only the read flags and nop ALU ops do not touch the
		  -- zero flag.
       if acode /= alu_rdf and acode /= alu_nop then
            if alu_result_int(7 downto 0) = data_zeros then
                z_int := '1';
            else
                z_int := '0';
            end if;
        end if;
        -- Output the ALU result and the new results of
        -- the Zero and Carry flag.
		  alu_result <= alu_result_int(7 downto 0);
		  zupdate <= z_int;
		  cupdate <= c_int;
	 end process alu;

	 -- The Conditional Jump Checker. This process determines if
	 -- a certain condition is met and sets the PC Load signal accordingly.
    jumpcheck : process (jcode, zflag, cflag) is
	 begin
        case jcode is
            when jump_jn =>   ldpc <= '0';
            when jump_jnz =>  ldpc <= not zflag;
            when jump_jz =>   ldpc <= zflag;
            when jump_jnc =>  ldpc <= not cflag;
            when jump_jc =>   ldpc <= cflag;
            when jump_jncz => ldpc <= not (cflag or zflag);
            when jump_jcz =>  ldpc <= cflag or zflag;
            when jump_j =>    ldpc <= '1';				
            when others =>    null;
		  end case;
	 end process jumpcheck;

    -- The Register File contains four registers and loads a
	 -- specified register if the enable is active.
    regfile : process (clk, areset, enr0, enr1, enr2, enr3) is
	 begin
        if areset = '1' then
            reg0 <= data_zeros;
            reg1 <= data_zeros;
            reg2 <= data_zeros;
            reg3 <= data_zeros;
        elsif rising_edge(clk) then
            if enr0 = '1' then
                reg0 <= alu_result;
            end if;
            if enr1 = '1' then
                reg1 <= alu_result;
            end if;
            if enr2 = '1' then
                reg2 <= alu_result;
            end if;
            if enr3 = '1' then
                reg3 <= alu_result;
            end if;
        end if;
    end process regfile;

	 -- Selection of the A input of the ALU
    muxa : process (sela, reg0, reg1, reg2, reg3) is
	 begin
        case sela is
		      when "00" => muxaout <= reg0;
		      when "01" => muxaout <= reg1;
		      when "10" => muxaout <= reg2;
		      when "11" => muxaout <= reg3;
            when others => null;
		  end case;
	 end process muxa;

	 -- Selection of the B input of the ALU
    muxb : process (selb, reg0, reg1, reg2, reg3, input, const) is
	 begin
        case selb is
		      when "000" => muxbout <= reg0;
		      when "001" => muxbout <= reg1;
		      when "010" => muxbout <= reg2;
		      when "011" => muxbout <= reg3;
		      when "100" => muxbout <= input;
		      when "101" => muxbout <= const;
		      --when "110" => muxaout <= ;
		      --when "111" => muxaout <= ;
				--when others => muxbout <= data_zeros;
				when others => muxbout <= (others => '-');
            --when others => null;
		  end case;
	 end process muxb;

	 -- Selection of the register to write back the result.
    decoderd : process (seld, en_d) is
	 begin
        enr0 <= '0';
        enr1 <= '0';
        enr2 <= '0';
        enr3 <= '0';
        if en_d = '1' then
            case seld is
                when "00" => enr0 <= '1';
                when "01" => enr1 <= '1';
                when "10" => enr2 <= '1';
                when "11" => enr3 <= '1';
                when others => null;
            end case;
        end if;
	 end process decoderd;

    -- The Program Counter. It can be loaded which
    -- effectively jumps to another program location.
    -- Program starts a address 0 on a reset.
    regpc : process (clk, areset) is
    variable pc_int : address_type;
	 begin
        if areset = '1' then
            pc_int := address_zeros;
        elsif rising_edge(clk) then
            if ldpc = '1' then
                pc_int := address;
            else
                pc_int := pc_int + 1;
            end if;
        end if;
        pc <= pc_int;
	 end process regpc;

    -- Instruction ROM decoding is nothing more then rewiring
    -- the instruction bits to the control inputs. Only VHDL
    -- makes such a trouble to do this the simple way...
    instr_rom : process (pc) is
	 begin
        sela    <= rom(to_integer(pc))(29 downto 28);
        selb    <= rom(to_integer(pc))(27 downto 25);
        seld    <= rom(to_integer(pc))(24 downto 23);
        acode   <= rom(to_integer(pc))(22 downto 19);
        const   <= unsigned(rom(to_integer(pc))(18 downto 11));
        en_d    <= rom(to_integer(pc))(10);
        jcode   <= rom(to_integer(pc))(9 downto 7);
        address <= unsigned(rom(to_integer(pc))(6 downto 0));
    end process instr_rom;

    -- The output is only here so that the synthesizer has
    -- something to synthesize.
    output <= reg0;
	 
end architecture rtl;
